/*-
 * Copyright (c) 1992 P. Keith Muller All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Keith Muller.
 * 4. The name of the developer may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * DEVELOPER (KEITH MULLER) BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if defined(LIBC_SCCS) && !defined(lint)
/* static char sccsid[] = "@(#)pax_beta	0.1 11/18/92"; */
#endif /* LIBC_SCCS and not lint */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include "pax.h"
#include "time_rng.h"
#include "extern.h"
static int str_sec();

/*
 * Routines to handle user selection of files based on the modification time
 * being within a specified time range (the non-standard -T flag). The user
 * may specify any number of different file modification time ranges. The ranges
 * are checked one at a time until a match is found (if at all).
 * If the file has a mtime that lies within one of the time ranges, the file
 * is selected. Time ranges may have a lower and/or a upper value. Ranges are
 * inclusive. If no time ranges are supplied to pax, all members in the archive
 * will be selected. If only a lower range is supplied, all files with a mtime
 * equal to or younger are selected. If only an upper range is supplied, all
 * files with a mtime equal to or older will be selected. When the lower value
 * is equal to the upper value, only files with a mtime of exactly that time
 * will be selected.
 */

static TIME_RNG *trhead = NULL;		/* time range list head */
static TIME_RNG *trtail = NULL;		/* time range list tail */

/*
 * trng_add()
 *	add a time range match to the time range list.
 *	This is a non-standard pax option. Lower and upper ranges are in the
 *	format: [yy[mm[dd[hh]]]]mm[.ss] and are comma separated.
 *	Time ranges are based on current time, so 1234 would specify a time of
 *	12:34 today.
 * Return:
 *	0 if the time range was added to the list, -1 otherwise
 */
int
trng_add(str)
	register char *str;
{
	register TIME_RNG *pt;
	register char *up_pt = NULL;
	register char *stpt;
	register int dot = 0;

	/*
	 * throw out the badly formed time ranges
	 */
	if ((str == NULL) || (*str == '\0')) {
		warn(1, "Empty time range string");
		return(-1);
	}
	for (stpt = str; *stpt != '\0'; ++stpt) {
		if ((*stpt >= '0') && (*stpt <= '9'))
			continue;
		if ((*stpt == ',') && (up_pt == NULL)) {
			*stpt = '\0';
			up_pt = stpt + 1;
			dot = 0;
			continue;
		}
		/*
		 * allow only one dot per range (secs)
		 */
		if ((*stpt == '.') && (!dot)) {
			++dot;
			continue;
		}
		warn(1, "Improperly specified time range: %s", str);
		return(-1);
	}
	/*
	 * allocate space for the time range and store the limits
	 */
	if ((pt = (TIME_RNG *)malloc(sizeof(TIME_RNG))) == NULL) {
		warn(1, "Unable to allocate storage for time range");
		return(-1);
	}
	pt->flags = 0;
	/*
	 * start off with the current time
	 */
	pt->low_time = pt->high_time = time((time_t *)NULL);
	if (*str != '\0') {
		/*
		 * add lower limit
		 */
		if (str_sec(str, &(pt->low_time)) < 0) {
			warn(1, "Illegal lower time range %s", str);
			(void)free((char *)pt);
			return(-1);
		}
		pt->flags |= HASLOW;
	}
	if ((up_pt != NULL) && (*up_pt != '\0')) {
		/*
		 * add upper limit
		 */
		if (str_sec(up_pt, &(pt->high_time)) < 0) {
			warn(1, "Illegal upper time range %s", up_pt);
			(void)free((char *)pt);
			return(-1);
		}
		pt->flags |= HASHIGH;
		/*
		 * check that the upper and lower do not overlap
		 */
		if (pt->flags & HASLOW) {
			if (pt->low_time > pt->high_time) {
				warn(1, "Upper %s and lower %s time overlap",
					up_pt, str);
				(void)free((char *)pt);
				return(-1);
			}
		}
	}
	pt->fow = NULL;
	if (trhead == NULL) {
		trtail = trhead = pt;
		return(0);
	}
	trtail->fow = pt;
	trtail = pt;
	return(0);
}

/*
 * trng_match()
 *	check if this files mtime falls within any supplied time range.
 * Return:
 *	0 if this archive member should be processed, 1 if it should be skipped
 */
int
trng_match(arcn)
	register ARCHD *arcn;
{
	register TIME_RNG *pt;

	/*
	 * no ranges specified, then all selected
	 */
	if (trhead == NULL)
		return(0);
	/*
	 * have to search down the list one at a time looking for a match.
	 * remember time range limits are inclusive.
	 */
	pt = trhead;
	while (pt != NULL) {
		if (pt->flags & HASLOW) {
			if (arcn->sb.st_mtime < pt->low_time) {
				pt = pt->fow;
				continue;
			}
		}
		if (pt->flags & HASHIGH) {
			if (arcn->sb.st_mtime > pt->high_time) {
				pt = pt->fow;
				continue;
			}
		}
		break;
	}
	if (pt == NULL)
		return(1);
	return(0);
}

/*
 * str_sec()
 *	Convert a time string in the format of [yy[mm[dd[hh]]]]mm[.ss] to gmt
 *	seconds. Tval already has current time loaded into it at entry.
 * Return:
 *	0 if converted ok, -1 otherwise
 */
static int
str_sec(str, tval)
	register char *str;
	time_t *tval;
{
	register struct tm *lt;
	register char *dot = NULL;

	lt = localtime(tval);
	if ((dot = index(str, '.')) != NULL) {
		/*
		 * seconds (.ss)
		 */
		*dot++ = '\0';
		if (strlen(dot) != 2)
			return(-1);
		if ((lt->tm_sec = ATOI2(dot)) > 61)
			return(-1);
	} else
		lt->tm_sec = 0;

	switch (strlen(str)) {
	case 10:
		/*
		 * year (yy)
		 * watch out for year 2000
		 */
		if ((lt->tm_year = ATOI2(str)) < 69)
			lt->tm_year += 100;
		str += 2;
		/* FALLTHROUGH */
	case 8:
		/*
		 * month (mm)
		 * watch out months are from 0 - 11 internally
		 */
		if ((lt->tm_mon = ATOI2(str)) > 12)
			return(-1);
		--lt->tm_mon;
		str += 2;
		/* FALLTHROUGH */
	case 6:
		/*
		 * day (dd)
		 */
		if ((lt->tm_mday = ATOI2(str)) > 31)
			return(-1);
		str += 2;
		/* FALLTHROUGH */
	case 4:
		/*
		 * hour (hh)
		 */
		if ((lt->tm_hour = ATOI2(str)) > 23)
			return(-1);
		str += 2;
		/* FALLTHROUGH */
	case 2:
		/*
		 * minute (mm)
		 */
		if ((lt->tm_min = ATOI2(str)) > 59)
			return(-1);
		break;
	default:
		return(-1);
	}
	/*
	 * convert broken-down time to GMT clock time seconds
	 */
	if ((*tval = mktime(lt)) == -1)
		return(-1);
	return(0);
}
